import numpy as np


# Classe de nombres duaux, AVEC DES TROUS.

# S'initialise en écrivant par exemple:
# u = Dual_Number( real = 2 , dual = 3) pour u = 2 + 3epsilon, ou plus simplement
# u = Dual_Number( 2 , 3 )
# Si u est un Dual_Number, u.r est sa partie réelle et u.d sa partie duale.
# L'addition et la multiplication peuvent combiner des int/float avec des Dual_Numbers.
# La soustraction et la division sont définies avec la négation simple et l'inverse multiplicatif
# en utilisant l'addition et la multiplication. Elles peuvent donc également combiner des
# int/floats avec des Dual_Numbers.


class Dual_Number:

    # Fonction d'initialisation de la classe.
    def __init__(self, real=1, dual=0):
        self.r = real
        self.d = dual

    # Fonction qui définit le string d'un nombre dual lorsqu'on le print.
    def __str__(self):
        if self.r == 0:
            if self.d == 0:
                return "0"
            else:
                return str(self.d) + "\u03B5"
        elif self.d == 0:
            return str(self.r)
        else:
            if self.d > 0:
                sym = "+"
            else:
                sym = "-"
            return str(self.r) + sym + str(abs(self.d)) + "\u03B5"

    ############### Surcharge des opérateurs ###################

    # Addition a.__add__(b)
    def __add__(a, b):
        # Cette partie assure que si l'on a un float ou un int, on le transforme en Dual_Number avant de faire l'opération,
        # que l'on doit alors faire avec les nombres duaux a et b2.
        if type(b) in [int, float]:
            b2 = Dual_Number(real=b)
        else:
            b2 = b
        # Fin de la partie pour assurer la compatibilité entre int/float et Dual_Numbers

        return Dual_Number(real=a.r + b2.r, dual=a.d + b2.d)

    # Addition a.__radd__(b)
    __radd__ = __add__

    # négation simple

    def __neg__(a):
        if type(a) in [int, float]:
            return -a
        else:
            return Dual_Number(real=-a.r, dual=-a.d)

    # soustraction a.__sub__(b)
    def __sub__(a, b):
        return a + b.__neg__()

    # soustraction a.__rsub__(b)
    def __rsub__(a, b):
        return b + a.__neg__()

    # multiplication a.__mul__(b)
    def __mul__(a, b):
        # Cette partie assure que si l'on a un float ou un int, on le transforme en Dual_Number avant de faire l'opération,
        # que l'on doit alors faire avec les nombres duaux a et b2.
        if type(b) in [int, float]:
            b2 = Dual_Number(real=b, dual=0)
        else:
            b2 = b
        # Fin de la partie pour assurer la compatibilité entre int/float et Dual_Numbers

        return Dual_Number(real=a.r * b2.r, dual=a.r * b2.d + a.d * b2.r)

    # multiplication a.__rmul__(b)
    __rmul__ = __mul__

    # definition de 1/a
    def mult_inverse(a):
        if type(a) in [int, float]:
            return 1/a
        elif a.r == 0:
            raise TypeError(
                "Division par un élément non-inversible :{}".format(a))
        else:
            return Dual_Number(real=1/a.r, dual=-a.d / (a.r * a.r))

    # division a/b : a.__truediv__(b)
    def __truediv__(a, b):
        return a * Dual_Number.mult_inverse(b)

    # division b/a : a.__rtruediv__(b)
    def __rtruediv__(a, b):
        return Dual_Number.mult_inverse(a) * b

    def exp(a):
        if type(a) in [int, float]:
            a2 = Dual_Number(real=a, dual=0)
        else:
            a2 = a
        return Dual_Number(real=np.exp(a2.r), dual=(np.exp(a2.r) * a2.d))

    def sin(a):
        if type(a) in [int, float]:
            a2 = Dual_Number(real=a, dual=0)
        else:
            a2 = a
        return Dual_Number(real=(np.sin(a2.r)), dual=(np.cos(a2.r) * a2.d))

    def cos(a):
        if type(a) in [int, float]:
            a2 = Dual_Number(real=a, dual=0)
        else:
            a2 = a
        return Dual_Number(real=(np.cos(a2.r)), dual=(-np.sin(a2.r) * a2.d))

    def log(a):
        """
        Natural log (ln)

        Returns:

        """
        if type(a) in [int, float]:
            a2 = Dual_Number(real=a, dual=0)
        else:
            a2 = a
        return Dual_Number(real=(np.log(a2.r)), dual=((1 / a2.r) * a2.d))


def Exercice1_a():
    a = Dual_Number(2, 3)
    b = Dual_Number(7, -2)

    print(f"{a} * {b}= {a * b}")


def Exercice1_b():
    a = Dual_Number(5, -2)
    b = Dual_Number(0, 3)
    c = Dual_Number(1, 1)

    print(f"{a} + {b} * {c} = {a + b * c}")


def Exercice1_c():
    a = Dual_Number(5, -2)
    b = Dual_Number(3, 1)

    print(f"{a} / {b}= {a / b}")


def Exercice1_d():
    a = Dual_Number(1, 1)
    b = Dual_Number(1, -1)
    c = Dual_Number(-1, 1)

    print(f"{a} / {b} * {c} = {a / b * c}")


def Test_exp_sin():
    a = Dual_Number(2, 1)

    print(f"exp(sin({a})) = {Dual_Number.exp(Dual_Number.sin(a))}")


def sigmoid_exo():
    def sigmoid(x): return 1 / (1 + Dual_Number.exp((-1) * x))

    [print(f"Sigmoid de {i} = {sigmoid(i)}") for i in range(10)]


def test():
    a = Dual_Number(2, 1)
    b = Dual_Number(2, -1)
    c = Dual_Number(1, 2)

    print(f"(a / b) * c = {a / b * c}")


def ex3():
    u = Dual_Number(-2, 9)
    v = Dual_Number(-22, 8)
    w = Dual_Number(3, -1)
    x = Dual_Number(9, -2)
    y = Dual_Number(-8, 10)

    print(f"v * ((u + w) / (x + y)) = {v * ((u + w) / (x + y))}")


if __name__ == "__main__":
    # test()
    # Exercice1_a()
    # Exercice1_b()
    # Exercice1_c()
    # Exercice1_d()
    # Test_exp_sin()
    # sigmoid_exo()
    ex3()
